import networkx as nx
from typing import List, Tuple
from geopy.distance import geodesic
from scipy.spatial import KDTree
import numpy as np
from itertools import combinations
from structs.landmarks import Landmark
from optimizer_v4 import print_res, link_list_simple, get_time
import os
import json
import heapq

# Heuristic function: distance to the goal
def heuristic(loc1: Tuple[float, float], loc2: Tuple[float, float], score2: int) -> float:
  return geodesic(loc1, loc2).meters



# A* planner to search through the graph
def a_star2(G, start_id, end_id, max_walking_time, must_do_nodes, max_landmarks, detour, speed):
  open_set = []
  heapq.heappush(open_set, (0, start_id, 0, [start_id], set([start_id])))
  best_path = None
  max_attractiveness = 0
  visited_must_do = set()

  while open_set:
    _, current_node, current_length, path, visited = heapq.heappop(open_set)

    # If current node is a must_do node and hasn't been visited yet, mark it as visited
    if current_node in must_do_nodes and current_node not in visited_must_do:
      visited_must_do.add(current_node)

    # Check if path includes all must_do nodes and reaches the end
    if current_node == end_id and all(node in visited for node in must_do_nodes):
      attractiveness = sum(G.nodes[node]['weight'] for node in path)
      if attractiveness > max_attractiveness:
        best_path = path
        max_attractiveness = attractiveness
      continue

    if len(path) > max_landmarks + 1:
      continue

    for neighbor in G.neighbors(current_node):
      if neighbor not in visited:
        #distance = int(geodesic(G.nodes[current_node]['pos'], G.nodes[neighbor]['pos']).meters * detour / (speed * 16.6666))
        distance = get_time(G.nodes[current_node]['pos'], G.nodes[neighbor]['pos'], detour, speed)
        if current_length + distance <= max_walking_time:
          new_path = path + [neighbor]
          new_visited = visited | {neighbor}
          estimated_cost = current_length + distance + get_time(G.nodes[neighbor]['pos'], G.nodes[end_id]['pos'], detour, speed)
          heapq.heappush(open_set, (estimated_cost, neighbor, current_length + distance, new_path, new_visited))

  # Check if all must_do_nodes have been visited
  if all(node in visited_must_do for node in must_do_nodes):
      return best_path, max_attractiveness
  else:
      return None, 0


def a_star(G, start_id, end_id, max_walking_time, must_do_nodes, max_landmarks, detour, speed):
    open_set = []
    heapq.heappush(open_set, (0, start_id, 0, [start_id], set([start_id])))
    best_path = None
    max_attractiveness = 0
    visited_must_do = set()

    while open_set:
        _, current_node, current_length, path, visited = heapq.heappop(open_set)

        # If current node is a must_do node and hasn't been visited yet, mark it as visited
        if current_node in must_do_nodes and current_node not in visited_must_do:
            visited_must_do.add(current_node)

        # Check if path includes all must_do nodes and reaches the end
        if current_node == end_id and all(node in visited for node in must_do_nodes):
            attractiveness = sum(G.nodes[node]['weight'] for node in path)
            if attractiveness > max_attractiveness:
                best_path = path
                max_attractiveness = attractiveness
            continue

        if len(path) > max_landmarks + 1:
            continue

        for neighbor in G.neighbors(current_node):
            if neighbor not in visited:
                distance = get_time(G.nodes[current_node]['pos'], G.nodes[neighbor]['pos'], detour, speed)
                if current_length + distance <= max_walking_time:
                    new_path = path + [neighbor]
                    new_visited = visited | {neighbor}
                    estimated_cost = current_length + distance + heuristic(G.nodes[neighbor]['pos'], G.nodes[end_id]['pos'], G.nodes[neighbor]['weight'])
                    heapq.heappush(open_set, (estimated_cost, neighbor, current_length + distance, new_path, new_visited))

    # Check if all must_do_nodes have been visited
    if all(node in visited_must_do for node in must_do_nodes):
        return best_path, max_attractiveness
    else:
        return None, 0
    


def find_path(G, start_id, finish_id, max_walking_time, must_do_nodes, max_landmarks) -> List[str]:
    
  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour = parameters['detour factor']
    speed = parameters['average walking speed']

  best_path, _ = a_star(G, start_id, finish_id, max_walking_time, must_do_nodes, max_landmarks, detour, speed)
  
  return best_path if best_path else []



# Function to dynamically adjust theta
def adjust_theta(num_nodes, theta_opt, target_ratio=2.0):
  # Start with an initial guess
  initial_theta = theta_opt
  # Adjust theta to aim for the target ratio of edges to nodes
  return initial_theta / (num_nodes ** (1 / target_ratio))


# Create a graph using NetworkX and generate the path without must_do
def generate_path(landmarks: List[Landmark], max_walking_time: float, max_landmarks: int) -> List[List[Landmark]]:

  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour = parameters['detour factor']
    speed = parameters['average walking speed']

  landmap = {}
  pos_dict = {}
  weight_dict = {}
  G = nx.Graph()

  # Add nodes to the graph with attractiveness
  for i, landmark in enumerate(landmarks):
    pos_dict[i] = landmark.location
    weight_dict[i] = landmark.attractiveness
    landmap[i] = landmark
    G.add_node(i, pos=landmark.location, weight=landmark.attractiveness)
    if landmark.name == 'start' :
      start_id = i
    elif landmark.name == 'finish' :
      finish_id = i

  coords = np.array(list(pos_dict.values()))
  kdtree = KDTree(coords)

  k = 5
  if len(landmarks) <= k :
     k = len(landmarks)-1
  
  for node, coord in pos_dict.items():
    indices = kdtree.query(coord, k + 1)[1]   # k+1 because the closest neighbor is the node itself
    for idx in indices[1:]:                   # skip the first one (itself)
      neighbor = list(pos_dict.keys())[idx]
      distance = get_time(coord, pos_dict[neighbor], detour, speed)
      G.add_edge(node, neighbor, weight=distance)
  
  print(f"Graph with {G.number_of_nodes()} nodes and {G.number_of_edges()} edges")
  print("Start computing path...")

  # Find the valid path using the greedy algorithm
  valid_path = find_path(G, start_id, finish_id, max_walking_time, [], max_landmarks)

  if not valid_path:
    return []  # No valid path found

  lis = [landmap[id] for id in valid_path]

  lis, _ = link_list_simple(lis)

  print_res(lis, len(landmarks))


  return lis



# Create a graph using NetworkX and generate the path
def generate_path2(landmarks: List[Landmark], max_walking_time: float, max_landmarks: int) -> List[List[Landmark]]:

  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour = parameters['detour factor']
    speed = parameters['average walking speed']


  landmap = {}
  pos_dict = {}
  weight_dict = {}
  must_do_nodes = []
  G = nx.Graph()
  # Add nodes to the graph with attractiveness
  for i, landmark in enumerate(landmarks):
    pos_dict[i] = landmark.location
    weight_dict[i] = landmark.attractiveness
    landmap[i] = landmark
    if landmark.must_do :
       must_do_nodes.append(i)
    G.add_node(i, pos=landmark.location, weight=landmark.attractiveness)
    if landmark.name == 'start' :
      start_id = i
    elif landmark.name == 'finish' :
      finish_id = i

  coords = np.array(list(pos_dict.values()))
  kdtree = KDTree(coords)

  k = 3
  for node, coord in pos_dict.items():
    indices = kdtree.query(coord, k + 1)[1]   # k+1 because the closest neighbor is the node itself
    for idx in indices[1:]:                   # skip the first one (itself)
      neighbor = list(pos_dict.keys())[idx]
      distance = get_time(coord, pos_dict[neighbor], detour, speed)
      G.add_edge(node, neighbor, weight=distance)



  # Add special edges between must_do nodes
  if len(must_do_nodes) > 0 :
    for node1, node2 in combinations(must_do_nodes, 2):
      if not G.has_edge(node1, node2):
        distance = get_time(G.nodes[node1]['pos'], G.nodes[node2]['pos'], detour, speed)
        G.add_edge(node1, node2, weight=distance)
  
  print(f"Graph with {G.number_of_nodes()} nodes and {G.number_of_edges()} edges")
  print("Computing path...")

  # Find the valid path using the greedy algorithm
  valid_path = find_path(G, start_id, finish_id, max_walking_time, must_do_nodes, max_landmarks)

  if not valid_path:
    return []  # No valid path found

  lis = [landmap[id] for id in valid_path]

  lis, tot_dist = link_list_simple(lis)

  print_res(lis, len(landmarks))


  return lis




def correct_path(tour: List[Landmark]) -> List[Landmark] :
  
  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour = parameters['detour factor']
    speed = parameters['average walking speed']

  G = nx.Graph()

  coords = []
  landmap = {}
  for i, landmark in enumerate(tour) :
    coords.append(landmark.location)
    landmap[i] = landmark
    G.add_node(i, pos=landmark.location, weight=landmark.attractiveness)

  kdtree = KDTree(coords)

  k = 3
  for node, coord in coords:
    indices = kdtree.query(coord, k + 1)[1]   # k+1 because the closest neighbor is the node itself
    for idx in indices[1:]:                   # skip the first one (itself)
      neighbor = list(coords)[idx]
      distance = get_time(coord, coords[neighbor], detour, speed)
      G.add_edge(node, neighbor, weight=distance)

  path = nx.approximation.traveling_salesman_problem(G, weight='weight', cycle=True)

  lis = [landmap[id] for id in path]

  lis, tot_dist = link_list_simple(lis)

  print_res(lis, len(tour))

  return path

