import numpy as np
import json, os

from typing import List, Tuple
from itertools import combinations
from scipy.optimize import linprog
from math import radians, sin, cos, acos
from shapely import Polygon
from geopy.distance import geodesic


from structs.landmarks import Landmark


# Function to print the result
def print_res(L: List[Landmark], L_tot):

    if len(L) == L_tot: 
        print('\nAll landmarks can be visited within max_steps, the following order is suggested : ')
    else :
        print('Could not visit all the landmarks, the following order is suggested : ')

    dist = 0
    for elem in L : 
        if elem.time_to_reach_next is not None :
            print('- ' + elem.name + ', time to reach next = ' + str(elem.time_to_reach_next))
            dist += elem.time_to_reach_next
        else : 
            print('- ' + elem.name)

    print("\nMinutes walked : " + str(dist))
    print(f"Visited {len(L)-2} out of {L_tot-2} landmarks")


# Function that returns the distance in meters from one location to another
def get_time(p1: Tuple[float, float], p2: Tuple[float, float], detour: float, speed: float) :
    
    # Compute the straight-line distance in m
    if p1 == p2 :
        return 0
    else: 
        #dist = 1000 * 6371.01 * acos(sin(radians(p1[0]))*sin(radians(p2[0])) + cos(radians(p1[0]))*cos(radians(p2[0]))*cos(radians(p1[1]) - radians(p2[1])))
        dist = geodesic(p1, p2).meters

    # Consider the detour factor for average city to determine walking distance (in m)
    walk_dist = dist*detour

    # Time to walk this distance (in minutes)
    walk_time = walk_dist/speed*(60/1000)

    """if walk_time > 15 :
        walk_time = 5*round(walk_time/5)
    else :
        walk_time = round(walk_time)"""
  

    return round(walk_time)


# Checks if the path is connected, returns a circle if it finds one and the RESULT
def is_connected(resx) -> bool:
    
    
    N = len(resx)               # length of res
    L = int(np.sqrt(N))         # number of landmarks. CAST INTO INT but should not be a problem because N = L**2 by def.
    resx = resx[:L*L]

    # first round the results to have only 0-1 values
    for i, elem in enumerate(resx):
        resx[i] = round(elem)
    
    n_edges = resx.sum()        # number of edges

    nonzeroind = np.nonzero(resx)[0] # the return is a little funny so I use the [0]

    nonzero_tup = np.unravel_index(nonzeroind, (L,L))

    ind_a = nonzero_tup[0].tolist()
    ind_b = nonzero_tup[1].tolist()

    edges = []
    edges_visited = []
    vertices_visited = []

    edge1 = (ind_a[0], ind_b[0])
    edges_visited.append(edge1)
    vertices_visited.append(edge1[0])

    for i, a in enumerate(ind_a) :
        edges.append((a, ind_b[i]))      # Create the list of edges

    remaining = edges
    remaining.remove(edge1)

    break_flag = False
    while len(remaining) > 0 and not break_flag:
        for edge2 in remaining :
            if edge2[0] == edge1[1] :
                if edge1[1] in vertices_visited :
                    edges_visited.append(edge2)
                    break_flag = True
                    break
                else :    
                    vertices_visited.append(edge1[1])
                    edges_visited.append(edge2)
                    remaining.remove(edge2)
                    edge1 = edge2

            elif edge1[1] == L-1 or edge1[1] in vertices_visited:
                        break_flag = True
                        break

    vertices_visited.append(edge1[1])


    if len(vertices_visited) == n_edges +1 :
      return vertices_visited, []
    else: 
      return vertices_visited, edges_visited


# Computes the time to reach from each landmark to the next
def link_list(order: List[int], landmarks: List[Landmark])->List[Landmark] :
    
  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour_factor = parameters['detour factor']
    speed = parameters['average walking speed']
  
  L =  []
  j = 0
  total_dist = 0
  while j < len(order)-1 :
    elem = landmarks[order[j]]
    next = landmarks[order[j+1]]

    d = get_time(elem.location, next.location, detour_factor, speed)
    elem.time_to_reach_next = d
    if elem.name not in ['start', 'finish'] :
      elem.must_do = True
    L.append(elem)
    j += 1
    total_dist += d

  L.append(next)

  return L, total_dist



# Constraint to respect only one travel per landmark. Also caps the total number of visited landmarks
def respect_number(L:int, A_ub, b_ub):

  ones = [1]*L
  zeros = [0]*L
  for i in range(L) :
    h = zeros*i + ones + zeros*(L-1-i) + [0]*(L-1)
    A_ub.append(h)
    b_ub.append(1)

  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    max_landmarks = parameters['max landmarks']

  A_ub.append(ones*L + [0]*(L-1))
  b_ub.append(max_landmarks+1)

  return A_ub, b_ub



def solve_optimizationv3(landmarks, max_walking_time):
  L = len(landmarks)

  # Read the parameters from the file
  with open (os.path.dirname(os.path.abspath(__file__)) + '/parameters/optimizer.params', "r") as f :
    parameters = json.loads(f.read())
    detour = parameters['detour factor']
    speed = parameters['average walking speed']

  # Create distance matrix
  A = np.zeros((L, L))
  for i in range(L):
    for j in range(L):
      if i != j:
        A[i, j] = get_time(landmarks[i].location, landmarks[j].location, detour, speed)

  # Define the linear program
  c = np.hstack((A.flatten(), [0]*(L-1)))
  bounds = [(0, 1) for _ in range(L*L+L-1)]

  # Flow conservation constraints
  A_eq = []
  b_eq = []

  # Each node (except start and end) has one incoming and one outgoing edge
  for i in range(L):
    if i == 0 or i == L-1:
      continue
    A_eq.append([1 if j // L == i else 0 for j in range(L*L)] + [0]*(L-1))
    b_eq.append(1)
    A_eq.append([1 if j % L == i else 0 for j in range(L*L)] + [0]*(L-1))
    b_eq.append(1)

  # Start node constraint
  A_eq.append([1 if j // L == 0 else 0 for j in range(L*L)] + [0]*(L-1))
  b_eq.append(1)

  # End node constraint
  A_eq.append([1 if j % L == L-1 else 0 for j in range(L*L)] + [0]*(L-1))
  b_eq.append(1)

  # Subtour elimination constraints
  A_ub = []
  b_ub = []

  # u_i - u_j + L*x_ij <= L-1 for all i != j
  for i in range(1, L):
    for j in range(1, L):
      if i != j:
        constraint = [0] * (L * L + L - 1)
        constraint[i * L + j] = L
        constraint[j * L + i] = -L
        A_ub.append(constraint)
        b_ub.append(L - 1)

  
  A_ub, b_ub = respect_number(L, A_ub, b_ub)                      # Respect max number of visits (no more possible stops than landmarks). 

  # Convert constraints to numpy arrays
  A_eq = np.array(A_eq)
  A_ub = np.array(A_ub)
  b_ub = np.array(b_ub)
  b_eq = np.array(b_eq)

  # Solve the linear program
  result = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq=b_eq, bounds=bounds, method='highs')

  if result.success:
    x = result.x[:L*L].reshape((L, L))
    path = []
    for i in range(L):
      for j in range(L):
        if x[i, j] > 0.5:
          path.append((i, j))
          print(f"({i}, {j})")

    order, _ = is_connected(result.x)
    L, _ = link_list(order, landmarks)
    
    print_res(L, len(landmarks))
    print("\nTotal score : " + str(int(-result.fun)))

    return L
  else:
    print("no results")
    return []



# Main optimization pipeline
# def solve_optimization (landmarks :List[Landmark], max_steps: int, printing_details: bool) :

#     L = len(landmarks)

#     # SET CONSTRAINTS FOR INEQUALITY
#     #c, A_ub, b_ub = init_ub_dist(landmarks, max_steps)              # Add the distances from each landmark to the other
#     A_ub, b_ub = respect_number(L, A_ub, b_ub)                      # Respect max number of visits (no more possible stops than landmarks). 
#     #A_ub, b_ub = break_sym(L, A_ub, b_ub)                           # break the 'zig-zag' symmetry
#     #A_ub, b_ub = prevent_subtours(L, A_ub, b_ub)

#     # SET CONSTRAINTS FOR EQUALITY
#     #A_eq, b_eq = init_eq_not_stay(L)                                # Force solution not to stay in same place
#     #A_eq, b_eq = respect_user_mustsee(landmarks, A_eq, b_eq)     # Check if there are user_defined must_see. Also takes care of start/goal
#     #A_eq, b_eq = respect_start_finish(L, A_eq, b_eq)                # Force start and finish positions
#     #A_eq, b_eq = respect_order(L, A_eq, b_eq)                       # Respect order of visit (only works when max_steps is limiting factor)

#     # SET BOUNDS FOR DECISION VARIABLE (x can only be 0 or 1)
#     x_bounds = [(0, 1)]*(L*L + L)

#     # Solve linear programming problem
#     res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)

#     # Raise error if no solution is found
#     if not res.success :
#             raise ArithmeticError("No solution could be found, the problem is overconstrained. Please adapt your must_dos")

#     # If there is a solution, we're good to go, just check for connectiveness
#     else :
#         order, circle = is_connected(res.x)
#         i = 0
#         timeout = 80
#         """while len(circle) != 0 and i < timeout:
#             A_ub, b_ub = prevent_config(res.x, A_ub, b_ub)
#             #A_ub, b_ub = break_cricle(order, len(landmarks), A_ub, b_ub)
#             res = linprog(c, A_ub=A_ub, b_ub=b_ub, A_eq=A_eq, b_eq = b_eq, bounds=x_bounds, method='highs', integrality=3)
#             if not res.success :
#                 raise ArithmeticError(f"No solution found after {timeout} iterations.")

#             order, circle = is_connected(res.x)
#             if len(circle) == 0 :
#                 break
#             print(i)
#             i += 1
        
#         if i == timeout :
#             raise TimeoutError(f"Optimization took too long. No solution found after {timeout} iterations.")
#         """
#         # Add the times to reach and stop optimizing
#         L, total_dist = link_list(order, landmarks)
            
#         if printing_details is True :
#             if i != 0 :
#                 print(f"Neded to recompute paths {i} times because of unconnected loops...")
#             print_res(L, len(landmarks))
#             print("\nTotal score : " + str(int(-res.fun)))

#         return L






